#!/usr/bin/python3
#
# Copyright 2022 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os
import re
import shutil
import subprocess
import sys

from pkg_resources import packaging


SOCKET = "unix:/run/podman/podman.sock"
PS_FORMAT = ('{"service": "{{.Names}}", "container": "{{.ID}}", '
             '"status": "{{.State}}", "healthy": "{{.Status}}"},')
BASE_FORMAT = "{service: .Name, container: .Id, status: .State.Running, "
RUNNING_REX = re.compile(r"Up .*(?P<health>\(u?n?healthy\))")
SKIP_LIST = ['.*_bootstrap', 'container-puppet-.*', '.*_db_sync',
             '.*_cron', 'create_.*_wrapper', '.*_wait_bundle',
             'configure_.*', '.*_fix_perms', '.*_init_logs?',
             '.*_init_perm', '(nova|placement)_wait_for_.*', 'nova_.*ensure_',
             '(swift_)?setup_.*', 'mysql_data_ownership', 'swift_copy_rings',
             'nova_statedir_owner']


class ExecuteError(Exception):
    def __init__(self, rc, msg):
        self.rc = rc
        self.msg = msg
        super().__init__()


def execute(cmd, workdir: str = None,
            prev_proc: subprocess.Popen = None) -> subprocess.Popen:
    # Note(mmagr): When this script is executed by collectd-sensubility started
    #              via collectd the script has non-root permission but inherits
    #              environment from collectd with root permission. We need
    #              to avoid sensubility access /root when using podman-remote.
    #              See https://bugzilla.redhat.com/show_bug.cgi?id=2091076 for
    #              more info.
    proc_env = os.environ.copy()
    proc_env["HOME"] = "/tmp"
    if type(cmd[0]) is list:  # multiple piped commands
        last = prev_proc
        for c in cmd:
            last = execute(c, workdir, last)
        return last
    else:  # single command
        inpipe = prev_proc.stdout if prev_proc is not None else None
        proc = subprocess.Popen(cmd, cwd=workdir, env=proc_env, stdin=inpipe,
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if prev_proc is not None:
            prev_proc.stdout.close()
            prev_proc.stderr.close()
        return proc


def get_health_from_status(status):
    healthy = status in ('healthy', 'running')
    return int(healthy)


def fetch_state_from_inspect(cont, fmt):
    proc = execute([
        [shutil.which('podman-remote'),
           '--url', SOCKET, 'inspect', cont["container"]],
        [shutil.which('jq'), '.[] | %s' % fmt]
    ])
    o, e = proc.communicate()
    if proc.returncode != 0:
        msg = "Failed to fetch status of %s: %s" % (cont, e.decode())
        raise ExecuteError(proc.returncode, msg)

    item = json.loads(o.decode())
    item['status'] = 'running' if item['status'] else 'stopped'
    if len(item['healthy']) > 0 and item['status'] != 'stopped':
        item['status'] = item['healthy']

    item['healthy'] = get_health_from_status(item['status'])
    return item


def fetch_state(item):
    if 'up ' in item['status'].lower():
        item['status'] = 'running'
    if item['status'] != 'running':
        item['status'] = 'stopped'

    match = RUNNING_REX.match(item['healthy'])
    if match:
        health = match.group('health')
    else:
        health = ''
    if health:
        item['status'] = health.strip("()")

    item['healthy'] = get_health_from_status(item['status'])
    return item


def fetch_container_health(ps_result, inspect_fmt):

    def container_filter(item):
        for rx in [re.compile(i) for i in SKIP_LIST]:
            if rx.match(item["service"]):
                return False
        return True
    try:
        containers = filter(container_filter, json.loads(ps_result))
    except json.decoder.JSONDecodeError:
        return 1, "%s\nFailed to parse above output." % ps_result

    out = []
    for cont in containers:
        try:
            out.append(fetch_state(cont))
        except Exception:
            # fallback to slow inspect way in case of any error
            try:
                out.append(fetch_state_from_inspect(cont, inspect_fmt))
            except ExecuteError as ex:
                return ex.rc, ex.msg
    return 0, out


if __name__ == "__main__":
    proc = execute([shutil.which('podman-remote'),
                    '--url', SOCKET, 'version',
                    '--format', r'{{.Server.Version}}'])
    o, e = proc.communicate()
    try:
        if packaging.version.parse(o.decode().strip()) >= packaging.version.parse("4.0.0"):
            inspect_fmt = BASE_FORMAT + "healthy: .State.Health.Status}"
        else:
            inspect_fmt = BASE_FORMAT + "healthy: .State.Healthcheck.Status}"
    except Exception:
        # keep podman-4.0.0+ format in case of version decoding error
        inspect_fmt = BASE_FORMAT + "healthy: .State.Health.Status}"

    proc = execute([shutil.which('podman-remote'), '--url', SOCKET,
                    'ps', '--all', '--format', PS_FORMAT])
    o, e = proc.communicate()
    if proc.returncode != 0:
        print("Failed to list containers:\n%s\n%s" % (o.decode(), e.decode()))
        sys.exit(1)

    rc, status = fetch_container_health("[%s]" % o.decode().strip("\n,"), inspect_fmt)
    if rc != 0:
        print("Failed to inspect containers:\n%s" % status)
        sys.exit(rc)
    print(json.dumps(status))



#!/usr/libexec/platform-python
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Migrate an undercloud's stack data to use ephemeral Heat. Queries for
existing stacks and exports necessary data from the stack to the default
consistent working directory before backing up and dropping the heat database.
"""

import argparse
import logging
import os
import subprocess
import tarfile
import tempfile
import time
import yaml

from heatclient.client import Client
import keystoneauth1
from mistralclient.api import base as mistralclient_exc
from mistralclient.api import client as mistral_client
import openstack
from tripleo_common.utils import plan as plan_utils


LOG = logging.getLogger('undercloud')
ROLE_DATA_MAP_FILE = ('/var/lib/tripleo-config/'
                      'overcloud-stack-role-data-file-map.yaml')


def parse_args():
    parser = argparse.ArgumentParser(
        description="Upgrade an undercloud for ephemeral Heat.")

    parser.add_argument(
        '--cloud', '-c',
        default='undercloud',
        help='The name of the cloud used for the OpenStack connection.')
    parser.add_argument(
        '--stack', '-s',
        action='append',
        help='The stack(s) to migrate to using ephemeral Heat. Can be '
             'specified multiple times. If not specified, all stacks '
             'will be migrated')
    parser.add_argument(
        '--working-dir', '-w',
        help='Directory to use for saving stack state. '
             'Defaults to ~/overcloud-deploy/<stack>')

    return parser.parse_args()


def database_exists():
    """Check if the heat database exists.

    :return: True if the heat database exists, otherwise False
    :rtype: bool
    """
    output = subprocess.check_output([
        'sudo', 'podman', 'exec', '-u', 'root', 'mysql',
        'mysql', '-e', 'show databases like "heat"'
    ])
    return 'heat' in str(output)


def backup_db(backup_dir):
    """Backup the heat database to the specified directory

    :param backup_dir: The directory to store the backup
    :type backup_dir: str
    :return: Database tarfile backup path
    :rtype: str
    """
    heat_dir = os.path.join(backup_dir, 'heat-launcher')
    if not os.path.isdir(heat_dir):
        os.makedirs(heat_dir)
    db_path = os.path.join(heat_dir, 'heat-db.sql')
    LOG.info("Backing up heat database to {}".format(db_path))
    with open(db_path, 'w') as out:
        subprocess.run([
            'sudo', 'podman', 'exec', '-u', 'root',
            'mysql', 'mysqldump', 'heat'], stdout=out,
            check=True)
    os.chmod(db_path, 0o600)

    tf_name = '{}-{}.tar.bzip2'.format(db_path, time.time())
    tf = tarfile.open(tf_name, 'w:bz2')
    tf.add(db_path, os.path.basename(db_path))
    tf.close()
    LOG.info("Created tarfile {}".format(tf_name))

    return tf_name


def _decode(encoded):
    """Decode a string into utf-8

    :param encoded: Encoded string
    :type encoded: string
    :return: Decoded string
    :rtype: string
    """
    if not encoded:
        return ""
    decoded = encoded.decode('utf-8')
    if decoded.endswith('\n'):
        decoded = decoded[:-1]
    return decoded


def _get_ctlplane_vip():
    """Get the configured ctlplane VIP

    :return: ctlplane VIP
    :rtype: string
    """
    return _decode(subprocess.check_output(
        ['sudo', 'hiera', 'controller_virtual_ip']))


def _get_ctlplane_ip():
    """Get the configured ctlplane IP

    :return: ctlplane IP
    :rtype: string
    """
    return _decode(subprocess.check_output(
        ['sudo', 'hiera', 'ctlplane']))


def _make_stack_dirs(stacks, working_dir):
    """Create stack directory if it does not already exist

    :stacks: List of overcloud stack names
    :type stacks: list
    :working_dir: Path to working directly
    :type working_dir: str
    :return: None
    """
    for stack in stacks:
        stack_dir = os.path.join(working_dir, stack)
        if not os.path.exists(stack_dir):
            os.makedirs(stack_dir)


def _log_and_raise(msg):
    """Log error message and raise Exception

    :msg: Message string that will be logged, and added in Exception
    :type msg: str
    :return: None
    """
    LOG.error(msg)
    raise Exception(msg)


def _get_role_data_file(heat, stack, fd, temp_file_path):
    """Get the role data file for a stack

    :param heat: Heat client
    :type heat: heatclient.client.Client
    :param stack: Stack name to query for passwords
    :type stack: str
    :fd: File descriptor
    :type fd: int
    :temp_file_path: Path to role data temp file
    :type temp_file_path: str
    :return: Path to the role data file
    :rtype:: str
    """
    try:
        _stack = heat.get_stack(stack)
        stack_outputs = {i['output_key']: i['output_value']
                         for i in _stack.outputs}
        roles_data = stack_outputs[
            'TripleoHeatTemplatesJinja2RenderingDataSources']['roles_data']

        with os.fdopen(fd, 'w') as tmp:
            tmp.write(yaml.safe_dump(roles_data))

        roles_data_file = temp_file_path
    except KeyError:
        if not os.path.isfile(ROLE_DATA_MAP_FILE):
            _log_and_raise("Overcloud stack role data mapping file: {} was "
                           "not found.".format(ROLE_DATA_MAP_FILE))

        with open(ROLE_DATA_MAP_FILE, 'r') as f:
            data = yaml.safe_load(f.read())

        roles_data_file = data.get(stack)
        if not roles_data_file or not os.path.isfile(roles_data_file):
            _log_and_raise("Roles data file: {} for stack {} not found."
                           .format(roles_data_file, stack))

    return roles_data_file


def drop_db():
    """Drop the heat database and heat users

    :return: None
    :rtype: None
    """
    LOG.info("Dropping Heat database")
    subprocess.check_call([
        'sudo', 'podman', 'exec', '-u', 'root',
        'mysql', 'mysql', 'heat', '-e',
        'drop database if exists heat'])
    LOG.info("Dropping Heat users")
    subprocess.check_call([
        'sudo', 'podman', 'exec', '-u', 'root',
        'mysql', 'mysql', '-e',
        'drop user if exists \'heat\'@\'{}\''.format(_get_ctlplane_ip())])
    subprocess.check_call([
        'sudo', 'podman', 'exec', '-u', 'root',
        'mysql', 'mysql', '-e',
        'drop user if exists \'heat\'@\'{}\''.format(_get_ctlplane_vip())])
    subprocess.check_call([
        'sudo', 'podman', 'exec', '-u', 'root',
        'mysql', 'mysql', '-e',
        'drop user if exists \'heat\'@\'%\''])


def export_passwords(heat, stack, stack_dir):
    """Export passwords from an existing stack and write them in Heat
    environment file format to the specified directory.

    :param cloud: Heat client
    :type cloud: heatclient.client.Client
    :param stack: Stack name to query for passwords
    :type stack: str
    :param stack_dir: Directory to save the generated Heat environment
        containing the password values.
    :type stack_dir: str
    :return: None
    :rtype: None
    """
    passwords_path = os.path.join(
        stack_dir, "tripleo-{}-passwords.yaml".format(stack))
    LOG.info("Exporting passwords for stack %s to %s"
             % (stack, passwords_path))
    passwords = plan_utils.generate_passwords(heat=heat, container=stack)
    password_params = dict(parameter_defaults=passwords)
    with open(passwords_path, 'w') as f:
        f.write(yaml.safe_dump(password_params))
    os.chmod(passwords_path, 0o600)


def export_networks(stack, stack_dir, cloud):
    """Export networks from an existing stack and write network data file.

    :param stack: Stack name to query for networks
    :type stack: str
    :param stack_dir: Directory to save the generated network data file
        containing the stack network definitions.
    :type stack_dir: str
    :param cloud: Cloud name used to send OpenStack commands
    :type cloud: str
    :return: None
    :rtype: None
    """
    network_data_path = os.path.join(
        stack_dir, "tripleo-{}-network-data.yaml".format(stack))
    LOG.info("Exporting network from stack %s to %s"
             % (stack, network_data_path))
    subprocess.check_call(['openstack', 'overcloud', 'network', 'extract',
                           '--stack', stack, '--output', network_data_path,
                           '--yes'], env={'OS_CLOUD': cloud})
    os.chmod(network_data_path, 0o600)


def export_network_virtual_ips(stack, stack_dir, cloud):
    """Export network virtual IPs from an existing stack and write network
    vip data file.

    :param stack: Stack name to query for networks
    :type stack: str
    :param stack_dir: Directory to save the generated data file
        containing the stack virtual IP definitions.
    :type stack_dir: str
    :param cloud: Cloud name used to send OpenStack commands
    :type cloud: str
    :return: None
    :rtype: None
    """
    vip_data_path = os.path.join(
        stack_dir, "tripleo-{}-virtual-ips.yaml".format(stack))
    LOG.info("Exporting network virtual IPs from stack %s to %s"
             % (stack, vip_data_path))
    subprocess.check_call(['openstack', 'overcloud', 'network', 'vip',
                           'extract', '--stack', stack, '--output',
                           vip_data_path, '--yes'], env={'OS_CLOUD': cloud})
    os.chmod(vip_data_path, 0o600)


def export_provisioned_nodes(heat, stack, stack_dir, cloud):
    """Export provisioned nodes from an existing stack and write baremetal
    deployment definition file.

    :param cloud: Heat client
    :type cloud: heatclient.client.Client
    :param stack: Stack name to query for networks
    :type stack: str
    :param stack_dir: Directory to save the generated data file
        containing the stack baremetal deployment definitions.
    :type stack_dir: str
    :param cloud: Cloud name used to send OpenStack commands
    :type cloud: str
    :return: None
    :rtype: None
    """
    fd, temp_file_path = tempfile.mkstemp()
    try:
        roles_data_file = _get_role_data_file(heat, stack, fd, temp_file_path)
        bm_deployment_path = os.path.join(
            stack_dir, "tripleo-{}-baremetal-deployment.yaml".format(stack))
        LOG.info("Exporting provisioned nodes from stack %s to %s"
                 % (stack, bm_deployment_path))
        subprocess.check_call(['openstack', 'overcloud', 'node', 'extract',
                               'provisioned', '--stack', stack, '--roles-file',
                               roles_data_file, '--output', bm_deployment_path,
                               '--working-dir', stack_dir, '--yes'], env={'OS_CLOUD': cloud})
        os.chmod(bm_deployment_path, 0o600)
    finally:
        os.remove(temp_file_path)


def main():
    logging.basicConfig()
    LOG.setLevel(logging.INFO)
    args = parse_args()

    sudo_user = os.environ.get('SUDO_USER')

    if not args.working_dir:
        if sudo_user:
            user_home = '~{}'.format(sudo_user)
        else:
            user_home = '~'

        working_dir = os.path.join(
            os.path.expanduser(user_home),
            'overcloud-deploy')
    else:
        working_dir = args.working_dir
    if not os.path.isdir(working_dir):
        os.makedirs(working_dir)

    conn = openstack.connection.from_config(cloud=args.cloud)
    try:
        heat = conn.orchestration
        _heatclient = Client('1', endpoint=conn.endpoint_for('orchestration'),
                             token=conn.auth_token)
    except keystoneauth1.exceptions.catalog.EndpointNotFound:
        LOG.error("No Heat endpoint found, won't migrate any "
                  "existing stack data.")
        raise

    try:
        stacks = args.stack or [s.name for s in heat.stacks()]
    except openstack.exceptions.HttpException:
        LOG.warning("No connection to Heat available, won't migrate any "
                    "existing stack data.")
        stacks = []

    # Make stack directories in the working directory if they don't not exist
    _make_stack_dirs(stacks, working_dir)

    for stack in stacks:
        stack_dir = os.path.join(working_dir, stack)
        export_networks(stack, stack_dir, args.cloud)
        export_network_virtual_ips(stack, stack_dir, args.cloud)
        export_provisioned_nodes(heat, stack, stack_dir, args.cloud)

    if database_exists():
        backup_dir = os.path.join(
            working_dir,
            'undercloud-upgrade-ephemeral-heat')
        db_tar_path = backup_db(backup_dir)
    else:
        LOG.warning("No database found to backup.")
        db_tar_path = None

    # Get and store ssh keys from mistral environment
    env_ssh_keys = None
    try:
        _workflowclient = mistral_client.client(
            mistral_url=conn.endpoint_for('workflow'),
            session=conn.session)
        env_ssh_keys = _workflowclient.environments.get('ssh_keys')
    except (keystoneauth1.exceptions.catalog.EndpointNotFound,
            mistralclient_exc.APIException):
        LOG.warning("Can not get ssh_keys from mistral environment"
                    "used for tripleo-admin user. This may cause "
                    "issues after upgrade.")
    for stack in stacks:
        stack_dir = os.path.join(working_dir, stack)
        if db_tar_path:
            # Symlink to the existing db backup
            os.symlink(db_tar_path,
                os.path.join(stack_dir, os.path.basename(db_tar_path)))
        export_passwords(_heatclient, stack, stack_dir)

        # Write the keys to stack_dir
        if env_ssh_keys:
            private_key = env_ssh_keys.variables['private_key']
            public_key = env_ssh_keys.variables['public_key']
            ssh_key_file = os.path.join(stack_dir, 'ssh_private_key')
            with os.fdopen(
                os.open(ssh_key_file,
                        flags=(os.O_WRONLY | os.O_CREAT | os.O_TRUNC),
                        mode=0o600), 'w') as fp:
                fp.write(private_key)

            with os.fdopen(
                os.open('{}.pub'.format(ssh_key_file),
                        flags=(os.O_WRONLY | os.O_CREAT | os.O_TRUNC),
                        mode=0o600), 'w') as fp:
                fp.write(public_key)

    if database_exists():
        drop_db()

    # Chown all files to original user if running under sudo
    if sudo_user:
        subprocess.run([
            'chown', '-R', '{}:{}'.format(sudo_user, sudo_user),
            working_dir],
            check=True)


if __name__ == '__main__':
    main()



#!/usr/bin/python3
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
import yaml

if len(sys.argv) != 3:
    raise RuntimeError('Not enough arguemnts')

FILE_A = sys.argv[1]
FILE_B = sys.argv[2]

with open(FILE_A, 'r') as file_a:
    a = yaml.safe_load(file_a)

with open(FILE_B, 'r') as file_b:
    b = yaml.safe_load(file_b)

if a != b:
    sys.exit("Files are different")

sys.exit(0)



#!/usr/bin/python3
#
# Copyright 2018 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import argparse
import logging
import os
import six
import socket
import sys
import time


from keystoneauth1 import loading
from keystoneauth1 import session

from novaclient import client

# In python3 SafeConfigParser was renamed to ConfigParser and the default
# for duplicate options default to true. In case of nova it is valid to
# have duplicate option lines, e.g. passthrough_whitelist which leads to
# issues reading the nova.conf
# https://bugs.launchpad.net/tripleo/+bug/1827775
if six.PY3:
    from six.moves.configparser import ConfigParser
    config = ConfigParser(strict=False)
else:
    from six.moves.configparser import SafeConfigParser
    config = SafeConfigParser()

debug = os.getenv('__OS_DEBUG', 'false')

if debug.lower() == 'true':
    loglevel = logging.DEBUG
else:
    loglevel = logging.INFO

logging.basicConfig(stream=sys.stdout, level=loglevel)
LOG = logging.getLogger('nova_wait_for_compute_service')

iterations = 60
timeout = 10
nova_cfg = '/etc/nova/nova.conf'

if __name__ == '__main__':
    parser = argparse.ArgumentParser(usage='%(prog)s [options]')
    parser.add_argument('-k', '--insecure',
                        action="store_false",
                        dest='insecure',
                        default=True,
                        help='Allow insecure connection when using SSL')

    args = parser.parse_args()
    LOG.debug('Running with parameter insecure = %s',
              args.insecure)

    if os.path.isfile(nova_cfg):
        try:
            config.read(nova_cfg)
        except Exception:
            LOG.exception('Error while reading nova.conf:')
    else:
        LOG.error('Nova configuration file %s does not exist', nova_cfg)
        sys.exit(1)

    my_host = config.get('DEFAULT', 'host')
    if not my_host:
        # If host isn't set nova defaults to this
        my_host = socket.gethostname()

    loader = loading.get_plugin_loader('password')
    auth = loader.load_from_options(
        auth_url=config.get('neutron',
                            'auth_url'),
        username=config.get('neutron',
                            'username'),
        password=config.get('neutron',
                            'password'),
        project_name=config.get('neutron',
                                'project_name'),
        project_domain_name=config.get('neutron',
                                       'project_domain_name'),
        user_domain_name=config.get('neutron',
                                    'user_domain_name'))
    sess = session.Session(auth=auth, verify=args.insecure)
    nova = client.Client('2.11', session=sess, endpoint_type='internal',
                         region_name=config.get('neutron', 'region_name'))

    # Wait until this host is listed in the service list
    for i in range(iterations):
        try:
            service_list = nova.services.list(binary='nova-compute')
            for entry in service_list:
                host = getattr(entry, 'host', '')
                zone = getattr(entry, 'zone', '')
                if host == my_host and zone != 'internal':
                    LOG.info('Nova-compute service registered')
                    sys.exit(0)
            LOG.info('Waiting for nova-compute service to register')
        except Exception:
            LOG.exception(
                'Error while waiting for nova-compute service to register')
        time.sleep(timeout)
sys.exit(1)

# vim: set et ts=4 sw=4 :



#!/usr/bin/python3
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import sys
import yaml


def filter_count(pair):
    key, _ = pair
    if key.endswith('Count'):
        return True
    else:
        return False


# We need to filter out roles with simillar name that our starts with picked up.
# But we also want to filter out baremetal node provision provided params, ContainerImage,Prepare and so on.
# Example to filter out:
# resource_registry:
#  OS::TripleO::ComputeHCI::Net::SoftwareConfig: OS::Heat::None
#  OS::TripleO::ComputeHCI::Ports::InternalApiPort: /usr/share/openstack-tripleo-heat-templates/network/ports/deployed_internal_api.yaml
#  OS::TripleO::ComputeHCI::Ports::StorageMgmtPort: /usr/share/openstack-tripleo-heat-templates/network/ports/deployed_storage_mgmt.yaml
#  OS::TripleO::ComputeHCI::Ports::StoragePort: /usr/share/openstack-tripleo-heat-templates/network/ports/deployed_storage.yaml
#  OS::TripleO::ComputeHCI::Ports::TenantPort: /usr/share/openstack-tripleo-heat-templates/network/ports/deployed_tenant.yaml
#  This one specifically as it's OSP17.1 EL8.4 compatibility:
#  OS::TripleO::ComputeHCI::Services::NovaLibvirt: /usr/share/openstack-tripleo-heat-templates/deployment/deprecated/nova/nova-libvirt-container-puppet.yaml
def filter_out_reg(pair):
    key, _ = pair
    if ("::Ports::" in key and key.endswith("Port")):
        return False
    elif key.endswith("Services::NovaLibvirt"):
        return False
    elif key.endswith("Net::SoftwareConfig"):
        return False
    else:
        return True


# parameter_defaults:
#  ComputeHCICount: 1
#  ComputeHCIHostnameFormat: computehci-%index%
#  ComputeHCINetworkConfigTemplate: /home/stack/overcloud-deploy/qe-Cloud-0/nic-configs/osdcompute.j2
#  ComputeHCIContainerImagePrepare:
def filter_out_f(pair, rolename=''):
    key, _ = pair
    if key.startswith(rolename):
        return False
    elif ("::Ports::" in key and key.endswith("Port:")):
        return False
    else:
        return True


def filter_out_f_2(pair, removeme=''):
    key, _ = pair
    return False if key.endswith(removeme) else True


def special_merge(dict_1, dict_2):
    dict_result = {**dict_1, **dict_2}
    for key, value in dict_result.items():
        if key in dict_1 and key in dict_2:
            if (type(dict_1[key]) is list):
                dict_result[key] = dict_1[key] + [value]
            elif (type(dict_1[key]) is dict):
                dict_result[key] = special_merge(dict_1[key], value)
            else:
                dict_result[key] = [value, dict_1[key]]
    return dict_result


def filter_role_parameters(env_files, rolename, new_role_env_file):
    n_role_params_def = {}
    n_role_res_registry = {}
    pd = {}
    rr = {}
    for file in env_files:
        if os.path.exists(file):
            with open(file, 'r') as env_file:
                data = yaml.safe_load(env_file)
                par = data.get('parameter_defaults', {})
                res = data.get('resource_registry', {})
                rr = special_merge(rr, res)
                pd = special_merge(pd, par)
    if pd:
        for key, value in pd.items():
            if key.startswith(rolename):
                n_role_params_def[key] = value
    if rr:
        for key, value in rr.items():
            if key.startswith("OS::TripleO::{}:".format(rolename)):
                n_role_res_registry[key] = value
    # We can have ComputeCount Compute1Count which
    # means we have picked on multiple roles and as we
    # cannot parse here for all role_specific tagged params
    # we simply pick on Count: to figure out if we don't happen
    # to have colision
    role_count = []
    filter_out = []
    role_count = dict(filter(filter_count, n_role_params_def.items()))
    if len(role_count) > 1:
        for key in role_count:
            if key != "{}Count".format(rolename):
                filter_out.append(key[:-5])
    for role in filter_out:
        n_role_params_def = dict(filter(
                                    lambda seq: filter_out_f(seq,
                                                             rolename=role),
                                    n_role_params_def.items()))
    for word in ['Count',
                 'HostnameFormat',
                 'NetworkConfigTemplate',
                 'ContainerImagePrepare',
                 'UpgradeInitCommand']:
        n_role_params_def = dict(filter(
                                   lambda seq: filter_out_f_2(seq,
                                                            removeme=word),
                                    n_role_params_def.items()))
    n_role_res_registry = dict(filter(
                                    lambda seq: filter_out_reg(seq),
                                    n_role_res_registry.items()))
    return (n_role_res_registry, n_role_params_def)


def rename_parameters(parameters, role_src, role_dst):
    rr, pd = parameters
    new_rr = {}
    new_pd = {}
    for registry in rr.keys():
        new_rr[registry.replace(role_src, role_dst)] = rr[registry]
    for parameter in pd.keys():
        new_pd[parameter.replace(role_src, role_dst)] = pd[parameter]
    return (new_rr, new_pd)


def write_parameters(parameters, new_role_env_file):
    rr, pd = parameters
    with open(new_role_env_file, 'w') as new_file:
        # I don't think we can blindly dump here, if one is empty we skip dumping
        dump_var = {}
        if (pd != {}):
            dump_var['parameter_defaults'] = pd
        if (rr != {}):
            dump_var['resource_registry'] = rr
        if (dump_var != {}):
            yaml.dump(dump_var,
                      new_file,
                      default_flow_style=False)
        else:
            dump_var['parameter_defaults'] = {}
            yaml.dump(dump_var,
                      new_file,
                      default_flow_style=False)


def parse_opts(argv):
    parser = argparse.ArgumentParser()

    parser.add_argument('--environment', '-e', dest='e',
                        help='envs', action = 'append')

    parser.add_argument('--output-file', '-o', dest='output_file', required=True,
                        help='Output file where the outcome is written')

    parser.add_argument('--rolename-src', dest='rolename_src', required=True,
                        help='The name of the role to copy')

    parser.add_argument('--rolename-dst', dest='rolename_dst', required=True,
                        help='The name of the role with paramaters copied')

    opts = parser.parse_known_args(argv[1:])
    return opts


opts = parse_opts(sys.argv)
parameters = filter_role_parameters(opts[0].e, opts[0].rolename_src, opts[0].output_file)
parameters = rename_parameters(parameters, opts[0].rolename_src, opts[0].rolename_dst)
write_parameters(parameters, opts[0].output_file)
